/*	$eterna: bozohttpd.c,v 1.29 2001/09/22 10:10:00 mrg Exp $	*/

/*
 * Copyright (c) 1997, 1998, 1999, 2000, 2001 Matthew R. Green
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer and
 *    dedication in the documentation and/or other materials provided
 *    with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/* this program is dedicated to the Great God of Processed Cheese */

/*
 * bozohttpd.c:  minimal httpd; provides only these features:
 *	- HTTP/0.9 (by virtue of ..)
 *	- HTTP/1.0
 *	- HTTP/1.1
 *	- CGI/1.1 this will only be provided for "system" scripts.
 *	- automatic "missing trailing slash" redirections.
 *	- configurable translation of /~user/ to ~user/public_html,
 *	  however, this does not include cgi-bin support (as a design
 *	  feature).
 *	- access lists via libwrap.
 *	- virtual hosting.
 *	- not that we do not even pretend to understand MIME, but
 *	  rely only on the HTTP specification.
 *	- ipv6 support (contributed by itojun@iijlab.net, KAME).
 *	- automatic `index.html' generation.
 *	- configurable server name.
 */

/*
 * requirements for minimal http/1.1 (at least, as documented in
 * <draft-ietf-http-v11-spec-rev-06> which expired may 18, 1999):
 *
 *	- 14.15: content-encoding handling. [1]
 *
 *	- 14.16: content-length handling.  this is only a SHOULD header
 *	  thus we could just not send it ever.  [1]
 *
 *	- 14.17: content-type handling. [1]
 *
 *	- 14.25/28: if-{,un}modified-since handling.  maybe do this.
 *
 * [1] need to revisit to ensure proper behaviour
 *
 * and the following is a list of features that we do not need
 * to have due to other limits, or are too lazy.  there are more
 * of these than are listed, but these are of particular note,
 * and could perhaps be implemented.
 *
 *	- 3.5/3.6: content/transfer codings.  probably can ignore
 *	  this?  we "SHOULD"n't.  but 4.4 says we should ignore a
 *	  `content-length' header upon reciept of a `transfer-encoding'
 *	  header.
 *
 *	- 5.1.1: request methods.  only MUST support GET and HEAD,
 *	  but there are new ones besides POST that are currently
 *	  supported: OPTIONS PUT DELETE TRACE and CONNECT, plus
 *	  extensions not yet known?
 *
 * 	- 10.1: we can ignore informational status codes
 *
 *	- 10.3.3/10.3.4/10.3.8:  just use '302' codes always.
 *
 *	- 14.1/14.2/14.3/14.27: we do not support Accept: headers..
 *	  just ignore them and send the request anyway.  they are
 *	  only SHOULD.
 *
 *	- 14.5/14.16/14.35: we don't do ranges.  from section 14.35.2
 *	  `A server MAY ignore the Range header'.  but it might be nice.
 *
 *	- 14.9: we aren't a cache.
 *
 *	- 14.15: content-md5 would be nice...
 *	
 *	- 14.24/14.26/14.27: be nice to support this...
 *
 *	- 14.44: not sure about this Vary: header.  ignore it for now.
 */

/*
 * optional features on a compile-time basis.
 */
#if 0
#undef NO_USER_SUPPORT		/* don't support /~user requests */
#undef NO_CGIBIN_SUPPORT	/* don't support cgi-bin requests */
#undef NO_DIRINDEX_SUPPORT	/* don't support directory indexing */
#undef NO_DAEMON_MODE		/* don't support daemon mode */
#undef DEBUG			/* include debugging support */
#undef SOCKOLD			/* use old socket api.  this will go away */
#endif

#ifdef SOCKOLD
# define NO_DAEMON_MODE		/* this needs new socket api */
#endif

#ifndef INDEX_HTML
#define INDEX_HTML	"index.html"
#endif
#ifndef PUBLIC_HTML
#define PUBLIC_HTML	"public_html"
#endif

/*
 * And so it begins ..
 */

#include <sys/param.h>
#include <sys/queue.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/wait.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#ifndef NO_CGIBIN_SUPPORT
#include <paths.h>
#endif /* NO_CGIBIN_SUPPORT */
#ifndef NO_USER_SUPPORT
#include <pwd.h>
#endif /* NO_USER_SUPPORT */
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#ifndef MAX_WAIT_TIME
#define	MAX_WAIT_TIME	60	/* hang around for 60 seconds max */
#endif

/* headers */
struct _headers {
	char	*header;
	char	*value;
	SIMPLEQ_ENTRY(_headers)	next;
};

/* http_req */
typedef struct {
	int	method;
#define	HTTP_GET	0x01
#define HTTP_POST	0x02
#define HTTP_HEAD	0x03
#define HTTP_OPTIONS	0x04	/* not supported */
#define HTTP_PUT	0x05	/* not supported */
#define HTTP_DELETE	0x06	/* not supported */
#define HTTP_TRACE	0x07	/* not supported */
#define HTTP_CONNECT	0x08	/* not supported */
	char	*methodstr;
	char	*url;
	char	*proto;
	char	*content_type;
	char	*content_length;
	char	*allow;
	char	*host;		/* HTTP/1.1 Host: */
	SIMPLEQ_HEAD(, _headers)	headers;
	int	nheaders;
} http_req;

/* debug flags */
#define DEBUG_LIGHT	0
#define DEBUG_NORMAL	1
#define DEBUG_FAT	2
#define DEBUG_OBESE	3

/* variables and functions */

#ifndef NO_CGIBIN_SUPPORT
static	char	*cgibin;	/* cgi-bin directory */

static	void	process_cgi __P((http_req *));
#endif /* NO_CGIBIN_SUPPORT */


#ifndef NO_USER_SUPPORT
static	int	uflag;		/* allow /~user/ translation */
static	char	*public_html	= PUBLIC_HTML;

static	char	*transform_user __P((http_req *, int *));
#endif /* NO_USER_SUPPORT */


#ifndef NO_DIRINDEX_SUPPORT
static	int	Xflag;		/* do directory indexing */

static	void	directory_index __P((http_req *, char *));
#endif /* NO_DIRINDEX_SUPPORT */


static	int	bflag;		/* background; drop into daemon mode. */
#ifndef NO_DAEMON_MODE
static	char	*iflag;		/* bind address; default INADDR_ANY. */
static	char	*Iflag = "http";/* bind port; default "http". */
static	int	sock = -1;	/* bound socket. */

static	void	sigchild __P((int));	/* SIGCHLD handler. */
#endif /* NO_DAEMON_MODE */


#ifdef DEBUG
static	int	dflag = 0;	/* debugging level */

static	void	debug__ __P((int, char *, ...));
#define debug(x)	debug__ x
#else
#define	debug(x)	
#endif /* DEBUG */


static	int	rflag;		/* redirect missing slash */
static	int	nflag;		/* avoid gethostby*() */
static	char	*myname;	/* my name */
static	char	*slashdir;	/* www slash directory */
static	char	*vpath;		/* virtual directory base */

static	char	date[40];	/* text date string */

static	char	*server_software = "bozotic HTTP server version 5.09";
static	char	*index_html	= INDEX_HTML;
static	char	http_09[]	= "HTTP/0.9";
static	char	http_10[]	= "HTTP/1.0";
static	char	http_11[]	= "HTTP/1.1";
static	char	text_plain[]	= "text/plain";

static	void	usage __P((void));
static	void	alarmer __P((int));
static	int	alarmhit;

static	void	parse_request __P((char *, char **, char **, char **));
static	http_req *read_request __P((void));
static	void	process_request __P((http_req *));
static	char	*transform_request __P((http_req *, int *));
static	void	handle_redirect __P((http_req *, char *));
static	void	print_header __P((http_req *, struct stat *, char *, char *));

static	void	check_virtual __P((http_req *));
static	void	fix_url_percent __P((http_req *));
static	void	process_method __P((http_req *, char *));
static	void	process_proto __P((http_req *, char *));
static	char	*content_type __P((http_req *, char *));
static	char	*content_encoding __P((http_req *, char *));

static	void	warning __P((char *, ...));
static	void	error __P((int, char *, ...));

static	void	http_error __P((int, http_req *, char *));
static	char	*http_errors_short __P((int));
static	char	*http_errors_long __P((int));

static	char	*dgetln __P((int, ssize_t *));
static	char	*strnsep __P((char **, char *, ssize_t *));

	int	main __P((int, char *[]));

static void
usage()
{
	extern	char *__progname;		/* XXX */

	warning("usage: %s [options] slashdir [myname]", __progname);
	warning("options:");
#ifndef NO_DIRINDEX_SUPPORT
	warning("\t-X\t\tenable automatic directory index support");
#endif
#ifdef DEBUG
	warning("\t-d\t\tenable debug support");
#endif
#ifndef NO_USER_SUPPORT
	warning("\t-u\t\tenable ~user/public_html support");
	warning("\t-p dir\t\tchange `public_html' directory name]");
#endif
#ifndef NO_CGIBIN_SUPPORT
	warning("\t-c cgibin\tenable cgi-bin support in this directory");
#endif
#ifndef NO_DAEMON_MODE
	warning("\t-b background and go into daemon mode");
	warning("\t-i address\tbind on this address (daemon mode only)");
	warning("\t-I port\tbind on this port (daemon mode only)");
#endif
	warning("\t-S version\tset server version string");
	warning("\t-v virtualroot\tenable virtual host support in this directory");
	warning("\t-x index\tchange default `index.html' file name");
	error(1, "%s failed to start", server_software);
}

int
main(argc, argv)
	int	argc;
	char	*argv[];
{
	http_req *http_request;
	struct	tm *tm;
	time_t	now;
	extern	char *__progname;		/* XXX */
	int	c;

	openlog(__progname, LOG_PID, LOG_FTP);

	while ((c = getopt(argc, argv, "I:S:Xbc:dhi:np:ruv:x:")) != -1) {
		switch(c) {
		case 'S':
			server_software = optarg;
			break;

		case 'X':
#ifndef NO_DIRINDEX_SUPPORT
			Xflag = 1;
			break;
#else
			error(1, "directory indexing is not enabled");
			/* NOTREACHED */
#endif /* NO_DIRINDEX_SUPPORT */

		case 'b':
#ifndef NO_DAEMON_MODE
			bflag = 1;
			break;
#else
			error(1, "Daemon mode is not enabled");
			/* NOTREACHED */
#endif /* NO_DAEMON_MODE */

		case 'c':
#ifndef NO_CGIBIN_SUPPORT
			cgibin = optarg;
			break;
#else
			error(1, "CGI is not enabled");
			/* NOTREACHED */
#endif /* NO_CGIBIN_SUPPORT */

		case 'd':
#ifdef DEBUG
			dflag++;
			break;
#else
			error(1, "Debugging is not enabled");
			/* NOTREACHED */
#endif /* DEBUG */

		case 'i':
#ifndef NO_DAEMON_MODE
			iflag = optarg;
			break;
#else
			error(1, "Daemon mode is not enabled");
			/* NOTREACHED */
#endif /* NO_DAEMON_MODE */

		case 'I':
#ifndef NO_DAEMON_MODE
			Iflag = optarg;
			break;
#else
			error(1, "Daemon mode is not enabled");
			/* NOTREACHED */
#endif /* NO_DAEMON_MODE */

		case 'n':
			nflag = 1;
			break;

		case 'p':
#ifndef NO_USER_SUPPORT
			public_html = optarg;
			break;
#else
			error(1, "User support is not enabled");
			/* NOTREACHED */
#endif /* NO_USER_SUPPORT */

		case 'r':
			rflag = 1;
			break;

		case 'u':
#ifndef NO_USER_SUPPORT
			uflag = 1;
			break;
#else
			error(1, "User support is not enabled");
			/* NOTREACHED */
#endif /* NO_USER_SUPPORT */

		case 'v':
			vpath = optarg;
			break;

		case 'x':
			index_html = optarg;
			break;
		default:
			usage();
			/* NOTREACHED */
		}
	}
	argc -= optind;
	argv += optind;

	if (argc == 1) {
		myname = malloc(MAXHOSTNAMELEN+1);
		if (myname == NULL)
			error(1, "malloc");
		/* XXX we do not check for FQDN here */
		if (gethostname(myname, MAXHOSTNAMELEN+1) < 0)
			error(1, "gethostname");
		myname[MAXHOSTNAMELEN] = '\0';
	} else if (argc == 2)
		myname = argv[1];
	else
		usage();
	
	slashdir = argv[0];
	debug((DEBUG_OBESE, "slashdir is %s", slashdir));

	/* Sun, 06 Nov 1994 08:49:37 GMT */
	now = time(NULL);
	tm = gmtime(&now);	/* HTTP/1.1 spec rev 06 sez GMT only */
	strftime(date, sizeof date, "%a, %d %b %Y %H:%M:%S GMT", tm);

#ifndef NO_DAEMON_MODE
	/*
	 * deal with daemon mode.
	 */
	if (bflag) {
		struct addrinfo h, *r;
		int s, e;

		daemon(1, 0);
		
		if (iflag == NULL)
			iflag = myname;
		memset(&h, 0, sizeof h);
		h.ai_family = PF_UNSPEC;
		h.ai_socktype = SOCK_STREAM;
		h.ai_flags = AI_PASSIVE;
		e = getaddrinfo(iflag, Iflag, &h, &r);
		if (e)
			error(1, "getaddrinfo([%s]:%s): %s", iflag, Iflag,
			    gai_strerror(e));
		if ((sock = socket(r->ai_family, SOCK_STREAM, 0)) == -1)
			error(1, "socket: %s", strerror(errno));
		if (bind(sock, r->ai_addr, r->ai_addrlen) == -1)
			error(1, "bind: %s", strerror(errno));
		if (listen(sock, SOMAXCONN) == -1)
			error(1, "socket: %s", strerror(errno));
		freeaddrinfo(r);

		signal(SIGCHLD, sigchild);
	}
#endif /* NO_DAEMON_MODE */

	/*
	 * read and process the HTTP request.
	 */
	do {
		if (http_request = read_request()) {
			process_request(http_request);
			return (0);
		}
	} while (bflag);

	return (0);
}

/*
 * convert "in" into the three parts of a request (first line)
 */
static void
parse_request(in, method, url, proto)
	char	*in;
	char	**method;
	char	**url;
	char	**proto;
{
	ssize_t	len;
	char	*val;
	
	*method = *url = *proto = NULL;		/* set them up */

	len = (ssize_t)strlen(in);
	val = strnsep(&in, " \t\n\r", &len);
	if (len < 1 || val == NULL)
		return;
	*method = val;
	while (*in == ' ' || *in == '\t')
		in++;
	val = strnsep(&in, " \t\n\r", &len);
	if (len < 1) {
		if (len == 0)
			*url = val;
		else
			*url = in;
		return;
	}
	*url = val;
	if (in) {
		while (*in && (*in == ' ' || *in == '\t'))
			in++;
		if (*in)
			*proto = in;
	}
}

/*
 * send a HTTP/1.1 408 response if we timeout.
 */
void
alarmer(sig)
	int sig;
{

	alarmhit = 1;
}

/*
 * This function reads a http request from stdin, returning a pointer to a
 * http_req structure, describing the request.
 */
http_req *
read_request()
{
	struct	sigaction	sa;
	char	*str, *val, *method, *url, *proto;
	ssize_t	len;
	int	line = 0;
	http_req *request;

#ifndef NO_DAEMON_MODE
	if (bflag) {
		struct sockaddr_storage ss;
		socklen_t slen;
		int fd;

		/*
		 * wait for a connection, then fork() and return NULL in the parent,
		 * who will come back here waiting for another connection.  read the
		 * request in in the child, and return it, for processing.
		 */
		fd = accept(sock, (struct sockaddr *)&ss, &slen);
		if (fd == -1) {
			if (errno != EAGAIN)
				error(1, "accept: %s", strerror(errno));
			return NULL;
		}
		switch (fork()) {
		case -1: /* eep, failure */
			warning("fork() failed, sleeping for 10 seconds: %s",
			    strerror(errno));
			sleep(10);
			return NULL;

		case 0: /* child */
			break;

		default: /* parent */
			close(fd);
			return NULL;
		}

		/* setup stdin/stdout/stderr */
		dup2(fd, 0);
		dup2(fd, 1);
		dup2(fd, 2);
		close(fd);
	}

#endif /* NO_DAEMON_MODE */

	request = malloc(sizeof *request);
	if (request == NULL)
		error(1, "malloc of request descriptor");
	request->allow = request->host = request->content_type =
	    request->content_length = NULL;

	/*
	 * setup a timer to make sure the request is not hung
	 */
	sa.sa_handler = alarmer;
	sigemptyset(&sa.sa_mask);
	sigaddset(&sa.sa_mask, SIGALRM);
	sa.sa_flags = 0;
	(void)sigaction(SIGALRM, &sa, NULL);	/* XXX */

	alarm(MAX_WAIT_TIME);
	while ((str = dgetln(STDIN_FILENO, (size_t *)&len)) != NULL) {
		alarm(0);
		if (alarmhit)
			http_error(408, NULL, "request timed out");
		line++;
		str = strdup(str);	/* we use this copy */
		if (line == 1) {
			if (len < 1)
				http_error(404, NULL, "null method");
			warning("got request ``%s''", str);
			debug((DEBUG_FAT, "read_req, getting request: ``%s''",
			    str));

			parse_request(str, &method, &url, &proto);
			if (method == NULL)
				http_error(404, NULL, "null method");
			if (url == NULL)
				http_error(404, NULL, "null url");

			/*
			 * note that we parse the proto first, so that we
			 * can more properly parse the method and the url.
			 */
			process_proto(request, proto);
			process_method(request, method);
			request->url = url;

			/* http/0.9 has no header processing */
			if (request->proto == http_09)
				break;
		} else {		/* incoming headers */
			struct	_headers *hdr;

			if (*str == '\0')
				break;

			val = strnsep(&str, ":", &len);
			debug((DEBUG_OBESE,
			    "read_req2: after strnsep: str ``%s'' val ``%s''",
			    str, val));
			if (val == NULL || len == -1)
				http_error(404, request, "no header");
			while (*str == ' ' || *str == '\t')
				len--, str++;
			
			hdr = malloc(sizeof *hdr);
			if (hdr == NULL)
				http_error(500, request, "malloc hdr");
			hdr->header = val;
			if (str && *str)
				hdr->value = str;
			else
				hdr->value = " ";

			SIMPLEQ_INSERT_TAIL(&request->headers, hdr, next);
			request->nheaders++;

			if (strcasecmp(hdr->header, "content-type") == 0)
				request->content_type = hdr->value;
			else if (strcasecmp(hdr->header, "content-length") == 0)
				request->content_length = hdr->value;
			else if (strcasecmp(hdr->header, "host") == 0)
				request->host = hdr->value;
			/* HTTP/1.1 rev06 draft spec: 14.20 */
			else if (strcasecmp(hdr->header, "expect") == 0)
				http_error(417, request, "we don't support Expect:");

			debug((DEBUG_FAT, "adding header %s: %s",
			    hdr->header, hdr->value));
		}
		alarm(MAX_WAIT_TIME);
	}

	/* now, clear it all out */
	alarm(0);
	(void)signal(SIGALRM, SIG_DFL);

	/* RFC1945, 8.3 */
	if (request->method == HTTP_POST && request->content_length == NULL)
		http_error(400, request, "missing content length");

	/* HTTP/1.1 draft rev-06, 14.23 & 19.6.1.1 */
	if (request->proto == http_11 && request->host == NULL)
		http_error(400, request, "missing Host header");

	debug((DEBUG_FAT, "read_request returns url %s in request", request->url));
	return (request);
}

/*
 * process_request does the following:
 *	- check the request is valid
 *	- process cgi-bin if necessary
 *	- transform a filename if necesarry
 *	- return the HTTP request
 */
void
process_request(request)
	http_req *request;
{
	struct	stat sb;
	char	*file, *type, *encoding;
	int	fd, isindex;

	/* note that transform_request chdir()'s if required */
	file = transform_request(request, &isindex);
	if (file == NULL)
		http_error(404, request, "empty file after transform");

	fd = open(file, O_RDONLY);
	if (fd < 0) {
		debug((DEBUG_FAT, "open failed: %s", strerror(errno)));
		if (errno == EPERM)
			http_error(403, request, "no permission to open file");
		else if (errno == ENOENT) {
#ifndef NO_DIRINDEX_SUPPORT
			if (isindex && Xflag) {
				file[strlen(file) - strlen(index_html)] = '\0';
				debug((DEBUG_FAT, "trying dir index (%s)",
				    file));
				directory_index(request, file);
				return;
			}
#endif /* NO_DIRINDEX_SUPPORT */
			http_error(404, request, "no file");
		} else
			http_error(500, request, "open file");
	}
	if (fstat(fd, &sb) < 0)
		http_error(500, request, "can't fstat");
	if (S_ISDIR(sb.st_mode))
		handle_redirect(request, file);
	/* XXX RFC1945 10.9 If-Modified-Since (http code 304) */

	(void)printf("%s 200 OK\r\n", request->proto);

	if (request->proto != http_09) {
		type = content_type(request, file);
		encoding = content_encoding(request, file);

		print_header(request, &sb, type, encoding);
		(void)putchar('\r');
		(void)putchar('\n');
	}
	(void)fflush(stdout);

	if (request->method != HTTP_HEAD) {
		void *addr, *oaddr;
		off_t sz = sb.st_size;

		oaddr = addr = mmap(0, (size_t)sz, PROT_READ,
		    MAP_SHARED, fd, 0);
		if (addr == (void *)-1)
			error(1, "mmap failed: %s", strerror(errno));

		if (madvise(addr, sz, MADV_SEQUENTIAL) < 0)
			warning("madvise failed, ignoring");

#define WRSZ	(64 * 1024)
		while (sz > WRSZ) {
			if (write(STDOUT_FILENO, addr, WRSZ) != WRSZ)
				error(1, "write failed: %s", strerror(errno));
			sz -= WRSZ;
			(char *)addr += WRSZ;
		}
		if (sz && write(STDOUT_FILENO, addr, sz) != sz)
			error(1, "final write failed: %s", strerror(errno));
		if (munmap(oaddr, (size_t)sb.st_size) < 0)
			warning("munmap failed");
	}
	(void)close(fd);
}

#ifndef NO_DIRINDEX_SUPPORT
/*
 * output a directory index.
 */
void
directory_index(request, dirname)
	http_req *request;
	char *dirname;
{
	struct stat sb;
	struct dirent *de;
	struct tm *tm;
	DIR *dp;
	char buf[MAXPATHLEN];
	int l;

	if (stat(dirname, &sb) < 0 ||
	    (dp = opendir(dirname)) == NULL) {
		if (errno == EPERM)
			http_error(403, request,
			    "no permission to open directory");
		else if (errno == ENOENT)
			http_error(404, request, "no file");
		else
			http_error(500, request, "open directory");
		return;
	}

	(void)printf("%s 200 OK\r\n", request->proto);

	if (request->proto != http_09) {
		print_header(request, &sb, "text/html", "");
		(void)putchar('\r');
		(void)putchar('\n');
	}
	(void)fflush(stdout);

	if (request->method == HTTP_HEAD) {
		closedir(dp);
		return;
	}

	(void)printf("<html><head><title>Index of %s</title></head>\r\n",
	    request->url);
	(void)printf("<body><h1>Index of %s</h1>\r\n", request->url);
	(void)printf("<pre>\r\n");
	(void)printf("Name                       "
	    "Last modified          "
	    "Size\n");
	(void)printf("<hr>\r\n\r\n");

	while ((de = readdir(dp)) != NULL) {
		if (strcmp(de->d_name, ".") == 0)
			continue;

		l = 0;

		if (strcmp(de->d_name, "..") == 0) {
			(void)printf("<a href=\"../\">");
			l += printf("Parent Directory");
		} else if (de->d_type == DT_DIR) {
			(void)printf("<a href=\"%s/\">", de->d_name);
			l += printf("%s/", de->d_name);
		} else {
			(void)printf("<a href=\"%s\">", de->d_name);
			l += printf("%s", de->d_name);
		}
		(void)printf("</a>");

		for (; l < 25; l++)
			(void)putchar(' ');

		(void)snprintf(buf, sizeof buf, "%s/%s", dirname,
		    de->d_name);
		if (stat(buf, &sb))
			(void)printf("?                         ?");
		else {
			tm = gmtime(&sb.st_mtime);
			(void)strftime(buf, sizeof buf, "%d-%b-%Y %R", tm);
			l += printf("%s", buf);

			for (; l < 45; l++)
				(void)putchar(' ');

			if (de->d_type == DT_REG)
				(void)printf("%7dkB",
				    (int)(sb.st_size >> 10));
			else
				(void)printf("        -");
		}
		(void)printf("\r\n");
	}

	closedir(dp);

	(void)printf("</pre><hr></body></html>\r\n");
}
#endif /* NO_DIRINDEX_SUPPORT */

#ifndef NO_DAEMON_MODE 
static	void
sigchild(signo)
	int signo;
{
	int status;

	while (waitpid(-1, &status, WNOHANG) != -1);
}
#endif /* NO_DAEMON_MODE */

/*
 * deal with virtual host names; we do this:
 *	if we have a virtual path root (vpath), and we are given a
 *	virtual host spec (Host: ho.st or http://ho.st/), see if this
 *	directory exists under vpath.  if it does, use this as the
 #	new slashdir.
 */
void
check_virtual(request)
	http_req *request;
{
	char *url = request->url, *s;
	struct dirent **list;
	size_t len;
	int i;

	if (!vpath)
		goto out;

	/*
	 * convert http://virtual.host/ to request->host
	 */
	debug((DEBUG_OBESE, "checking for http:// virtual host in ``%s''", url));
	if (strncasecmp(url, "http://", 7) == 0) {
		/* we would do virtual hosting here? */
		url += 7;
		s = strchr(url, '/');
		/* HTTP/1.1 draft rev-06, 5.2: URI takes precedence over Host: */
		request->host = url;
		if (s) {
			request->url = strdup(s);
			*s = 0;
		} else
			request->url = strdup("/");
		debug((DEBUG_OBESE, "got host ``%s'' url is now ``%s''", request->host, request->url));
	} else if (!request->host)
		goto out;


	/*
	 * ok, we have a virtual host, use scandir(3) to find a case
	 * insensitive match for the virtual host we are asked for.
	 * note that if the virtual host is the same as the master,
	 * we don't need to do anything special.
	 */
	len = strlen(request->host);
	debug((DEBUG_OBESE, "check_virtual: checking host ``%s'' under vpath ``%s'' for url ``%s''", request->host, vpath, request->url));
	if (strncasecmp(myname, request->host, len) != 0) {
		s = 0;
		for (i = scandir(vpath, &list, 0, 0); i--; list++) {
			debug((DEBUG_OBESE, "looking at dir``%s''", (*list)->d_name));
			if (strncasecmp((*list)->d_name, request->host,
			    len) == 0) {
				/* found it, punch it */
				myname = (*list)->d_name;
				if (asprintf(&s, "%s/%s", vpath, myname) < 0)
					error(1, "vasprintf");
				break;
			}
		}
		if (s == 0)
			http_error(404, request, "unknown URL");
	} else
out:
		s = slashdir;

	/*
	 * ok, nailed the correct slashdir, chdir to it
	 */
	if (chdir(s) < 0)
		error(1, "can't chdir %s: %s", s, strerror(errno));
}

/*
 * transform_request does this:
 *	- ``expand'' %20 crapola
 *	- punt if it doesn't start with /
 *	- look for "http://myname/" and deal with it.
 *	- check if /cgi-bin/ and call process_cgi() if so
 *	- check for ~user and call transform_user() if so
 *	- if the length > 1, check for trailing slash.  if so,
 *	  add the index.html file
 *	- if the length is 1, return the index.html file
 *	- disallow anything ending up with a file starting
 *	  at "/" or having ".." in it.
 *	- anything else is a really weird internal error
 */
char *
transform_request(request, isindex)
	http_req *request;
	int *isindex;
{
	char	*file, *url;
	size_t	len;

	file = NULL;	/* XXX gcc */
	*isindex = 0;
	debug((DEBUG_FAT, "tf_req: url %s", request->url));
	fix_url_percent(request);
	check_virtual(request);
	url = request->url;
	if (url[0] != '/') {
		http_error(404, request, "unknown URL");
#ifndef NO_CGIBIN_SUPPORT
	} else if (strncmp(url + 1, "cgi-bin/", 8) == 0) {
		debug((DEBUG_FAT, "calling process_cgi"));
		process_cgi(request);
#endif /* NO_CGIBIN_SUPPORT */
#ifndef NO_USER_SUPPORT
	} else if ((len = strlen(url)) > 1 && uflag && url[1] == '~') {
		if (url[2] == '\0')
			http_error(404, request, "missing username");
		if (strchr(url + 2, '/') == NULL)
			handle_redirect(request, url);
		debug((DEBUG_FAT, "calling transform_user"));
		return (transform_user(request, isindex));
	} else if (len > 1) {
#else
	} else if ((len = strlen(url)) > 1) {
#endif /* NO_USER_SUPPORT */
		debug((DEBUG_FAT, "url[len-1] == %c", url[len-1]));
		if (url[len-1] == '/') {	/* append index.html */
			*isindex = 1;
			debug((DEBUG_FAT, "appending index.html"));
			file = malloc(len + strlen(index_html) + 1);
			if (file == NULL)
				http_error(500, request, "malloc or file copy");
			strcpy(file, url + 1);
			strcat(file, index_html);
		} else
			file = strdup(url + 1);
	} else if (len == 1) {
		debug((DEBUG_OBESE, "tf_req: len == 1"));
		file = strdup(index_html);
		*isindex = 1;
	} else		/* len == 0 ? */
		http_error(500, request, "request->url is nul?");

	/*
	 * look for "http://myname/" and deal with it as necessary.
	 */

	/*
	 * stop traversing outside our domain 
	 *
	 * XXX true security only comes from our parent using chroot(2)
	 * before execve(2)'ing us.
	 */
	if (*file == '/' || strstr(file, ".."))
		http_error(403, request, "illegal request");

	debug((DEBUG_FAT, "transform_request returned: %s", file));
	return (file);
}

#ifndef NO_USER_SUPPORT
/*
 * transform_user does this:
 *	- chdir's /~user/public_html
 *	- returns the rest of the file, index.html appended if required
 *
 * transform_request() is supposed to check that we have user support
 * enabled.
 */
char *
transform_user(request, isindex)
	http_req *request;
	int *isindex;
{
	char	c, *s, *file = NULL;
	struct	passwd *pw;

	*isindex = 0;

	if ((s = strchr(request->url + 2, '/')) != NULL) {
		*s++ = '\0';
		c = s[strlen(s)-1];
		*isindex = (c == '/' || c == '\0');
	}

	debug((DEBUG_FAT, "looking for user %s", request->url + 2));
	pw = getpwnam(request->url + 2);
	/* fix this up immediately */
	if (s)
		s[-1] = '/';
	if (pw == NULL)
		http_error(404, request, "no such user");

	debug((DEBUG_OBESE, "user %s home dir %s uid %d gid %d", pw->pw_name,
	    pw->pw_dir, pw->pw_uid, pw->pw_gid));

	if (chdir(pw->pw_dir) < 0) {
		warning("chdir1 error: %s: %s", pw->pw_dir, strerror(errno));
		http_error(403, request, "can't chdir to homedir");
	}
	if (chdir(public_html) < 0) {
		warning("chdir2 error: %s: %s", public_html, strerror(errno));
		http_error(403, request, "can't chdir to public_html");
	}
	if (s == NULL || *s == '\0')
		file = strdup(index_html);
	else {
		file = malloc(strlen(s) +
		    (*isindex ? strlen(index_html) + 2 : 1));
		if (file == NULL)
			http_error(500, request, "malloc failed");
		strcpy(file, s);
		if (*isindex) {
			/* XXX What is this for? */
			/* strcat(file, "/"); */
			strcat(file, index_html);
		}
	}

	/* see above */
	if (*file == '/' || strstr(file, ".."))
		http_error(403, request, "illegal request");

	debug((DEBUG_FAT, "transform_user returning %s under %s", file,
	    pw->pw_dir));
	return (file);
}
#endif /* NO_USER_SUPPORT */

/*
 * This is a message obtained somewhere from the Net by my wife.
 */
static	char redirect_msg[] =
"
<body>
<p>The directory you specified <b>%s</b> was not followed by a ``/''

<p>You asked our server for a URL that matched a directory here, but which
does not end in a slash (``/''). Many servers handle this case by telling
your client to retry the request with a URL that has a slash at the end.
However, this has bad performance effects on our server, and we would
rather display this message so that you, the user, can edit your ``hot
list'' such that the URLs you use to reach us always have trailing slashes
unless they are explicit files.

<p>If this seems like a horrible kludge, we agree, and we suggest that you
take your complaint up with the HTTP protocol designers. You can send us
mail if you wish, but be warned that we are relatively set in our ways on
this topic.
</body>
";

/*
 * If we are allowing automatic redirection, just do it, else return the
 * canned message, or the provided one.
 */
void
handle_redirect(request, file)
	http_req *request;
	char	*file;
{
	
	warning("redirecting %s", request->url);
	debug((DEBUG_FAT, "redirecting %s", request->url));
	if (rflag)
		(void)printf("%s 301 Document Moved\r\n", request->proto);
	else
		(void)printf("%s 501 Not Implemented\r\n", request->proto);
	if (request->proto != http_09) 
		print_header(request, NULL, "text/html", NULL);
	if (rflag) {
		if (request->proto != http_09)
			(void)printf("Location: http://%s%s/\r\n", myname,
			    request->url);
		(void)putchar('\r');
		(void)putchar('\n');
		if (request->method == HTTP_HEAD)
			goto head;
		(void)printf("<html><head><title>Document Moved</title></head>\n");
		(void)printf("<body><h1>Document Moved</h1>\n");
		(void)printf("This document had moved <a href=\"http://%s%s/\">here</a>\n",
		    myname, request->url);
		(void)printf("</body></html>\n");
	} else {
		(void)putchar('\r');
		(void)putchar('\n');
		if (request->method == HTTP_HEAD)
			goto head;
		(void)printf("<html><head><title>No trailing slash in URL</title></head>\n");
		(void)printf(redirect_msg, file);
		(void)printf("</html>\n");
	}
head:
	fflush(stdout);
	exit(0);
}

/* generic header printing routine */
void
print_header(request, sbp, type, encoding)
	http_req *request;
	struct stat *sbp;
	char *type;
	char *encoding;
{

	(void)printf("Date: %s\r\n", date);
	(void)printf("Server: %s\r\n", server_software);
	if (sbp) {
		char filedate[40];
		struct	tm *tm;

		tm = gmtime(&sbp->st_mtime);
		strftime(filedate, sizeof filedate,
		    "%a, %d %b %Y %H:%M:%S GMT", tm);
		(void)printf("Last-Modified: %s\r\n", filedate);
	}
	if (type && *type)
		(void)printf("Content-Type: %s\r\n", type);
	if (encoding && *encoding)
		(void)printf("Content-Encoding: %s\r\n", encoding);
	if (sbp)
		(void)printf("Content-Length: %qd\r\n", (long long)sbp->st_size);
	if (request && request->proto == http_11)
		(void)printf("Connection: close\r\n");
}

#ifndef NO_CGIBIN_SUPPORT
/*
 * Checks if the request has asked for a cgi-bin.  If this functionality has
 * been enabled, try to run the program, or return an error if it fails.
 */
void
process_cgi(request)
	http_req *request;
{
	struct	_headers *headp;
#ifndef SOCKOLD
	struct	sockaddr_storage ss;
#else
	struct	sockaddr ss;
#endif
	char	*host, *addr, *port, *info, *type, *clen, *query, *s, *t,
		    *path, *env, *command = NULL,
		    *url = strdup(request->url + strlen("/cgi-bin"));
	char	**envp, *argv[3];
	char	bufport[10];
	int	envpsize, ix, slen = sizeof(ss);

	debug((DEBUG_FAT, "process_cgi:  writing HTTP header .."));

	if (request->method == HTTP_HEAD)
		goto print_cgi_header;

	if ((s = strchr(url + 1, '/')) != NULL) {
		command = malloc(s - url);
		if (command == NULL)
			http_error(500, request, "malloc failed");
		strncpy(command, url + 1, s - url - 1);
		command[s - url - 1] = '\0';
		url = info = s;
	} else
		info = "";

	if (command == NULL)
		command = url;

	if ((s = strchr(url, '?')) != NULL) {
		*s++ = '\0';
		query = s;
	} else
		query = NULL;

	debug((DEBUG_FAT, "process_cgi: query %s cmd %s info %s", query,
	    command, info));

	type = request->content_type;
	clen = request->content_length;

	if (getpeername(0, (struct sockaddr *)&ss, &slen) < 0)
		host = addr = NULL;
	else {
#ifndef SOCKOLD
		char	hbuf[NI_MAXHOST], abuf[NI_MAXSERV];

		addr = host = NULL;
		if (getnameinfo((struct sockaddr *)&ss, slen,
		    abuf, sizeof abuf, NULL, 0, NI_NUMERICHOST) == 0)
			addr = abuf;
		if (nflag != 0 && getnameinfo((struct sockaddr *)&ss, slen,
		    hbuf, sizeof hbuf, NULL, 0, 0) == 0)
			host = hbuf;
#else
		struct	hostent *hp;
		struct	sockaddr_in *sin;

		switch (((struct sockaddr *)&ss)->sa_family) {
		case AF_INET:
			sin = (struct sockaddr_in *)&ss;
			addr = inet_ntoa(sin->sin_addr);
			if (nflag == 0 ||
			    (hp = gethostbyaddr((char *)&sin->sin_addr,
			     sizeof(sin->sin_addr), AF_INET)) == NULL)
				host = NULL;
			else
				host = hp->h_name;
			break;
		default:
			host = addr = NULL;
			break;
		}
#endif
	}

	slen = sizeof(ss);
	if (getsockname(0, (struct sockaddr *)&ss, &slen) < 0)
		port = NULL;
	else {
#ifndef SOCKOLD
		if (getnameinfo((struct sockaddr *)&ss, slen, NULL, 0,
		    bufport, sizeof bufport, NI_NUMERICSERV) == 0)
			port = bufport;
		else
			port = NULL;
#else
		(void)snprintf(port = bufport, sizeof bufport, "%d",
		    ntohs(((struct sockaddr_in *)&ss)->sin_port));
#endif
	}

	envpsize = 10 + request->nheaders + 
	    (info && *info ? 1 : 0) +
	    (query && query ? 1 : 0) +
	    (type && *type ? 1 : 0) +
	    (clen && *clen ? 1 : 0) +
	    (host && *host ? 1 : 0) +
	    (addr && *addr ? 1 : 0) +
	    (port && *port ? 1 : 0);

	envp = malloc(sizeof(*envp) * envpsize);
	if (envp == NULL)
		http_error(500, request, "malloc failed");
	for (ix = 0; ix < envpsize; ix++)
		envp[ix] = NULL;
	ix = 0;

/* NOTE: not "side effects" safe! */
#define spsetenv(env, val) 					\
do {								\
	char	*s = malloc(strlen(env) + strlen(val) + 2);	\
								\
	if (s == NULL)						\
		http_error(500, request, "malloc failed");	\
	strcpy(s, env); strcat(s, "="); strcat(s, val);		\
	debug((DEBUG_OBESE, "spsetenv: %s", s));		\
	envp[ix++] = s;						\
} while (0)

	for (headp = request->headers.sqh_first; headp;
	    headp = headp->next.sqe_next) {
		env = malloc(6 + strlen(headp->header) + 1 +
		    strlen(headp->value));
		if (env == NULL)
			http_error(500, request, "malloc failed");

		t = env;
		strcpy(t, "HTTP_");
		t += strlen(t);
		for (s = headp->header; *s; t++, s++)
			if (islower(*s))
				*t = toupper(*s);
			else
				*t = *s;
		debug((DEBUG_OBESE, "setting header %s as %s = %s",
		    headp->header, env, headp->value));
		spsetenv(env, headp->value);
		free(env);
	}
		
	spsetenv("PATH", _PATH_DEFPATH);
	spsetenv("IFS", " \t\n");
	spsetenv("SERVER_NAME", myname);
	spsetenv("GATEWAY_INTERFACE", "CGI/1.1");
	spsetenv("SERVER_PROTOCOL", request->proto);
	spsetenv("REQUEST_METHOD", request->methodstr);
	spsetenv("SCRIPT_NAME", command);
	spsetenv("SERVER_SOFTWARE", server_software);
	if (query && *query)
		spsetenv("QUERY_STRING", query);
	if (info && *info)
		spsetenv("PATH_INFO", info);
	if (type && *type)
		spsetenv("CONTENT_TYPE", type);
	if (clen && *clen)
		spsetenv("CONTENT_LENGTH", clen);
	if (port && *port)
		spsetenv("SERVER_PORT", port);
	if (host && *host)
		spsetenv("REMOTE_HOST", host);
	if (addr && *addr)
		spsetenv("REMOTE_ADDR", addr);
#undef spsetenv

	path = malloc(strlen(cgibin) + 1 + strlen(command) + 1);
	if (path == NULL)
		http_error(500, request, "malloc failed");
	strcpy(path, cgibin);
	strcat(path, "/");
	strcat(path, command);

	/* may as well wait as long as possible */
print_cgi_header:
	(void)printf("%s 200 Here it is\r\n", request->proto);
	if (request->proto != http_09)
		print_header(NULL, NULL, NULL, NULL);
	(void)fflush(stdout);
	if (request->method == HTTP_HEAD)
		exit(0);

	argv[0] = command;
	argv[1] = query;
	argv[2] = NULL;
	debug((DEBUG_FAT, "going exec %s %s %s", path, command, query));
	if (execve(path, argv, envp) < 0)
		error(1, "child exec failed");
}
#endif /* NO_CGIBIN_SUPPORT */

/* this fixes the %HH hack that RFC1738 requires.  */
static void
fix_url_percent(request)
	http_req *request;
{
	char	*s, *t, buf[3];

	/* fast forward to the first % */
	if ((s = strchr(request->url, '%')) == NULL)
		return;

	t = s;
	do {
		debug((DEBUG_OBESE, "fu_%%: got s == %%, s[1]s[2] == %c%c",
		    s[1], s[2]));
		if (s[1] == '\0' || s[2] == '\0')
			http_error(400, request,
			    "percent hack missing two chars afterwards");
		buf[0] = *++s;
		buf[1] = *++s;
		buf[2] = '\0';
		s++;
		*t = (char)strtol(buf, NULL, 16);
		debug((DEBUG_OBESE, "fu_%%: strtol put %c into *t", *t));
		if (*t++ == '\0')
			http_error(400, request, "percent hack got a 0 back");

		while (*s && *s != '%')
			*t++ = *s++;
	} while (*s);
	*t = '\0';
	debug((DEBUG_FAT, "fix_url_percent returns %s in url", request->url));
}

/*
 * process each type of HTTP method, setting this HTTP requests
 # method type.
 */
static struct method_map {
	char	*name;
	int	type;
} method_map[] = {
	{ "GET", 	HTTP_GET, },
	{ "POST",	HTTP_POST, },
	{ "HEAD",	HTTP_HEAD, },
#if 0	/* other non-required http/1.1 methods */
	{ "OPTIONS",	HTTP_OPTIONS, },
	{ "PUT",	HTTP_PUT, },
	{ "DELETE",	HTTP_DELETE, },
	{ "TRACE",	HTTP_TRACE, },
	{ "CONNECT",	HTTP_CONNECT, },
#endif
	{ NULL,		0, },
};

static void
process_method(request, method)
	http_req *request;
	char	*method;
{
	struct	method_map *mmp;

	for (mmp = method_map; mmp->name; mmp++)
		if (strcasecmp(method, mmp->name) == 0) {
			request->method = mmp->type;
			request->methodstr = mmp->name;
			return;
		}

	if (request->proto == http_11)
		request->allow = "GET, HEAD, POST";
	http_error(404, request, "unknown method");
}

/*
 * as the prototype string is not constant (eg, "HTTP/1.1" is equivalent
 * to "HTTP/001.01"), we MUST parse this.
 */
static void
process_proto(request, proto)
	http_req *request;
	char	*proto;
{
	char	*majorstr, *minorstr;
	int	majorint, minorint;

	if (proto == NULL) {
got_proto_09:
		request->proto = http_09;
		debug((DEBUG_OBESE, "request %s is http/0.9", request->url));
		return;
	}

	if (strncasecmp(proto, "HTTP/", 5) != 0)
		goto bad;
	majorstr = proto + 5;
	minorstr = strchr(majorstr, '.');
	if (minorstr == NULL)
		goto bad;
	*minorstr++ = 0;

	majorint = atoi(majorstr);
	minorint = atoi(minorstr);

	switch (majorint) {
	case 0:
		if (minorint != 9)
			break;
		goto got_proto_09;
	case 1:
		if (minorint == 0)
			request->proto = http_10;
		else if (minorint == 1)
			request->proto = http_11;
		else
			break;

		debug((DEBUG_OBESE, "request %s is %s", request->url,
		    request->proto));
		SIMPLEQ_INIT(&request->headers);
		request->nheaders = 0;
		return;
	}
bad:
	http_error(404, NULL, "unknown prototype");
}

/*
 * this map and the functions below map between filenames and the
 * content type and content encoding definitions.  this should become
 * a configuration file, perhaps like apache's mime.types (but that
 * has less info per-entry).
 */
static struct content_map {
	char	*name;		/* postfix of file */
	char	*type;		/* matching content-type */
	char	*encoding;	/* matching content-encoding */
	char	*encoding11;	/* matching content-encoding (HTTP/1.1) */
} content_map[] = {
	{ ".html",	"text/html",			"",		"", },
	{ ".htm",	"text/html",			"",		"", },
	{ ".gif",	"image/gif",			"",		"", },
	{ ".jpeg",	"image/jpeg",			"",		"", },
	{ ".jpg",	"image/jpeg",			"",		"", },
	{ ".jpe",	"image/jpeg",			"",		"", },
	{ ".png",	"image/png",			"",		"", },
	{ ".swf",	"application/x-shockwave-flash","",		"", },
	{ ".dcr",	"application/x-director",	"",		"", },
	{ ".pac",	"application/x-ns-proxy-autoconfig", "",	"", },
	{ ".pa",	"application/x-ns-proxy-autoconfig", "",	"", },
	{ ".tar",	"multipart/x-tar",		"",		"", },
	{ ".gtar",	"multipart/x-gtar",		"",		"", },
	{ ".tar.Z",	"multipart/x-tar",		"x-compress",	"compress", },
	{ ".tar.gz",	"multipart/x-tar",		"x-gzip",	"gzip", },
	{ ".taz",	"multipart/x-tar",		"x-gzip",	"gzip", },
	{ ".tgz",	"multipart/x-tar",		"x-gzip",	"gzip", },
	{ ".tar.z",	"multipart/x-tar",		"x-pack",	"x-pack", },
	{ ".Z",		"application/x-compress",	"x-compress",	"compress", },
	{ ".gz",	"application/x-gzip",		"x-gzip",	"gzip", },
	{ ".z",		"unknown",			"x-pack",	"x-pack", },
	{ ".hqx",	"application/mac-binhex40",	"",		"", },
	{ ".cpt",	"application/mac-compactpro",	"",		"", },
	{ ".doc",	"application/msword",		"",		"", },
	{ ".bin",	"application/octet-stream",	"",		"", },
	{ ".dms",	"application/octet-stream",	"",		"", },
	{ ".lha",	"application/octet-stream",	"",		"", },
	{ ".lzh",	"application/octet-stream",	"",		"", },
	{ ".exe",	"application/octet-stream",	"",		"", },
	{ ".class",	"application/octet-stream",	"",		"", },
	{ ".oda",	"application/oda",		"",		"", },
	{ ".pdf",	"application/pdf",		"",		"", },
	{ ".ai",	"application/postscript",	"",		"", },
	{ ".eps",	"application/postscript",	"",		"", },
	{ ".ps",	"application/postscript",	"",		"", },
	{ ".ppt",	"application/powerpoint",	"",		"", },
	{ ".rtf",	"application/rtf",		"",		"", },
	{ ".bcpio",	"application/x-bcpio",		"",		"", },
	{ ".vcd",	"application/x-cdlink",		"",		"", },
	{ ".cpio",	"application/x-cpio",		"",		"", },
	{ ".csh",	"application/x-csh",		"",		"", },
	{ ".dir",	"application/x-director",	"",		"", },
	{ ".dxr",	"application/x-director",	"",		"", },
	{ ".dvi",	"application/x-dvi",		"",		"", },
	{ ".hdf",	"application/x-hdf",		"",		"", },
	{ ".cgi",	"application/x-httpd-cgi",	"",		"", },
	{ ".skp",	"application/x-koan",		"",		"", },
	{ ".skd",	"application/x-koan",		"",		"", },
	{ ".skt",	"application/x-koan",		"",		"", },
	{ ".skm",	"application/x-koan",		"",		"", },
	{ ".latex",	"application/x-latex",		"",		"", },
	{ ".mif",	"application/x-mif",		"",		"", },
	{ ".nc",	"application/x-netcdf",		"",		"", },
	{ ".cdf",	"application/x-netcdf",		"",		"", },
	{ ".patch",	"application/x-patch",		"",		"", },
	{ ".sh",	"application/x-sh",		"",		"", },
	{ ".shar",	"application/x-shar",		"",		"", },
	{ ".sit",	"application/x-stuffit",	"",		"", },
	{ ".sv4cpio",	"application/x-sv4cpio",	"",		"", },
	{ ".sv4crc",	"application/x-sv4crc",		"",		"", },
	{ ".tar",	"application/x-tar",		"",		"", },
	{ ".tcl",	"application/x-tcl",		"",		"", },
	{ ".tex",	"application/x-tex",		"",		"", },
	{ ".texinfo",	"application/x-texinfo",	"",		"", },
	{ ".texi",	"application/x-texinfo",	"",		"", },
	{ ".t",	"application/x-troff",			"",		"", },
	{ ".tr",	"application/x-troff",		"",		"", },
	{ ".roff",	"application/x-troff",		"",		"", },
	{ ".man",	"application/x-troff-man",	"",		"", },
	{ ".me",	"application/x-troff-me",	"",		"", },
	{ ".ms",	"application/x-troff-ms",	"",		"", },
	{ ".ustar",	"application/x-ustar",		"",		"", },
	{ ".src",	"application/x-wais-source",	"",		"", },
	{ ".zip",	"application/zip",		"",		"", },
	{ ".au",	"audio/basic",			"",		"", },
	{ ".snd",	"audio/basic",			"",		"", },
	{ ".mpga",	"audio/mpeg",			"",		"", },
	{ ".mp2",	"audio/mpeg",			"",		"", },
	{ ".aif",	"audio/x-aiff",			"",		"", },
	{ ".aiff",	"audio/x-aiff",			"",		"", },
	{ ".aifc",	"audio/x-aiff",			"",		"", },
	{ ".ram",	"audio/x-pn-realaudio",		"",		"", },
	{ ".rpm",	"audio/x-pn-realaudio-plugin",	"",		"", },
	{ ".ra",	"audio/x-realaudio",		"",		"", },
	{ ".wav",	"audio/x-wav",			"",		"", },
	{ ".pdb",	"chemical/x-pdb",		"",		"", },
	{ ".xyz",	"chemical/x-pdb",		"",		"", },
	{ ".ief",	"image/ief",			"",		"", },
	{ ".tiff",	"image/tiff",			"",		"", },
	{ ".tif",	"image/tiff",			"",		"", },
	{ ".ras",	"image/x-cmu-raster",		"",		"", },
	{ ".pnm",	"image/x-portable-anymap",	"",		"", },
	{ ".pbm",	"image/x-portable-bitmap",	"",		"", },
	{ ".pgm",	"image/x-portable-graymap",	"",		"", },
	{ ".ppm",	"image/x-portable-pixmap",	"",		"", },
	{ ".rgb",	"image/x-rgb",			"",		"", },
	{ ".xbm",	"image/x-xbitmap",		"",		"", },
	{ ".xpm",	"image/x-xpixmap",		"",		"", },
	{ ".xwd",	"image/x-xwindowdump",		"",		"", },
	{ ".txt",	"text/plain",			"",		"", },
	{ ".rtx",	"text/richtext",		"",		"", },
	{ ".tsv",	"text/tab-separated-values",	"",		"", },
	{ ".etx",	"text/x-setext",		"",		"", },
	{ ".sgml",	"text/x-sgml",			"",		"", },
	{ ".sgm",	"text/x-sgml",			"",		"", },
	{ ".mpeg",	"video/mpeg",			"",		"", },
	{ ".mpg",	"video/mpeg",			"",		"", },
	{ ".mpe",	"video/mpeg",			"",		"", },
	{ ".qt",	"video/quicktime",		"",		"", },
	{ ".mov",	"video/quicktime",		"",		"", },
	{ ".avi",	"video/x-msvideo",		"",		"", },
	{ ".movie",	"video/x-sgi-movie",		"",		"", },
	{ ".ice",	"x-conference/x-cooltalk",	"",		"", },
	{ ".wrl",	"x-world/x-vrml",		"",		"", },
	{ ".vrml",	"x-world/x-vrml",		"",		"", },
	{ NULL,		NULL,				NULL,		NULL, },
};

/*
 * given the file name, return a valid Content-Type: value.
 */
char *
content_type(request, file)
	http_req *request;
	char	*file;
{
	size_t	len = strlen(file), nlen;
	struct	content_map	*map;

	for (map = content_map; map->name; map++) {
		nlen = strlen(map->name);
		if (nlen > len || strcmp(map->name, file + (len - nlen)) != 0)
			continue;
		return (map->type);
	}
	return (text_plain);
}

/*
 * given the file name, return a valid Content-Encoding: value.
 */
char *
content_encoding(request, file)
	http_req *request;
	char	*file;
{
	size_t	len = strlen(file), nlen;
	struct	content_map	*map;

	for (map = content_map; map->name; map++) {
		nlen = strlen(map->name);
		if (nlen > len || strcmp(map->name, file + (len - nlen)) != 0)
			continue;
		return (request->proto == http_11 ?
		    map->encoding11 : map->encoding);
	}
	return (NULL);
}

/*
 * XXX
 */
#ifdef DEBUG
void
debug__(int level, char *fmt, ...)
{
	va_list	ap;
	int savederrno;
	
	/* only log if the level is low enough */
	if (level < dflag)
		return;

	savederrno = errno;
	va_start(ap, fmt);
	vsyslog(LOG_DEBUG, fmt, ap);
	va_end(ap);
	errno = savederrno;
}
#endif /* DEBUG */

/* these are like warn() and err(), except for syslog not stderr */
void
warning(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vsyslog(LOG_INFO, fmt, ap);
	va_end(ap);
}

void
error(int code, char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vsyslog(LOG_INFO, fmt, ap);
	va_end(ap);
	exit(code);
}

/* the follow functions and variables are used in handling HTTP errors */
void
http_error(code, request, log)
	int	code;
	http_req *request;
	char	*log;
{
	static	char buf[BUFSIZ];	/* static so we don't need to malloc? */
	char	*header = http_errors_short(code);
	char	*reason = http_errors_long(code);
	char	*proto = (request && request->proto) ? request->proto : http_11;
	int	size;

	debug((DEBUG_FAT, "http_error %d: %s", code, log));
	if (header == NULL || reason == NULL)
		error(1, "http_errors*() failed (short = %p, long = %p)",
		    header, reason);

	if (request && request->url) {
		size = snprintf(buf, sizeof buf,
		    "<html><head><title>%s</title></head>\n"
		    "<body><h1>%s</h1>\n"
		    "%s: <pre>%s</pre>\n"
		    "<hr><address><a href=\"http://%s/\">%s</a></address>\n"
		    "</body></html>\n",
		    header, header, request->url, reason, myname, myname);
		if (size >= sizeof buf)
			warning("http_error buffer too small, truncated");
	} else
		size = 0;

	(void)printf("%s %s\r\n", proto, header);
	(void)printf("Content-type: text/html\r\n");
	(void)printf("Context-length: %d\r\n", size);
	(void)printf("Server: %s\r\n", server_software);
	if (request && request->allow)
		(void)printf("Allow: %s\r\n", request->allow);
	(void)putchar('\r');
	(void)putchar('\n');
	if (size)
		(void)puts(buf);
	fflush(stdout);

	exit(1);
}

/* short map between error code, and short/long messages */
static struct errors_map {
	int	code;			/* HTTP return code */
	char	*shortmsg;		/* short version of message */
	char	*longmsg;		/* long version of message */
} errors_map[] = {
	{ 400,	"400 Bad Request",	"The request was not valid", },
	{ 403,	"403 Forbidden",	"Access to this item has been denied", },
	{ 404, 	"404 Not Found",	"This item has not been found", },
	{ 408, 	"408 Request Timeout",	"This request took too long", },
	{ 417,	"417 Expectation Failed","Expectations not available", },
	{ 500,	"500 Internal Error",	"An error occured on the server", },
	{ 501,	"501 Not Implemented",	"This request is not available", },
	{ 0,	NULL,			NULL, },
};

static char *help = "DANGER! WILL ROBINSON! DANGER!";

static char *
http_errors_short(code)
	int	code;
{
	struct errors_map *ep;

	for (ep = errors_map; ep->code; ep++)
		if (ep->code == code)
			return (ep->shortmsg);
	return (help);
}

static char *
http_errors_long(code)
	int	code;
{
	struct errors_map *ep;

	for (ep = errors_map; ep->code; ep++)
		if (ep->code == code)
			return (ep->longmsg);
	return (help);
}

/* Below are various modified libc functions */

/*
 * returns -1 in lenp if the string ran out before finding a delimiter,
 * but is otherwise the same as strsep.  Note that the length must be
 * correctly passed in.
 */
char *
strnsep(strp, delim, lenp)
	char	**strp;
	char	*delim;
	ssize_t	*lenp;
{
	char	*s;
	const	char *spanp;
	int	c, sc;
	char	*tok;

	if ((s = *strp) == NULL)
		return (NULL);
	for (tok = s;;) {
		if (lenp && --(*lenp) == -1)
			return (NULL);
		c = *s++;
		spanp = delim;
		do {
			if ((sc = *spanp++) == c) {
				if (c == 0)
					s = NULL;
				else
					s[-1] = '\0';
				*strp = s;
				return (tok);
			}
		} while (sc != 0);
	}
	/* NOTREACHED */
}

/*
 * inspired by fgetln(3), but works for fd's.  should work identically
 * except it, however, does *not* return the newline, and it does nul
 * terminate the string.
 */
static char *
dgetln(fd, lenp)
	int	fd;
	ssize_t	*lenp;
{
	static	char *buffer;
	static	ssize_t buflen = 0;
	ssize_t	len;
	int	got_cr = 0;
	char	c, *nbuffer;

	/* initialise */
	if (buflen == 0) {
		buflen = 128;	/* should be plenty for most requests */
		buffer = malloc(buflen);
		if (buffer == NULL) {
			buflen = 0;
			return NULL;
		}
	}
	len = 0;

	/*
	 * we *have* to read one byte at a time, to not break cgi
	 * programs (for we pass stdin off to them).  could fix this
	 * by becoming a fd-passing program instead of just exec'ing
	 * the program
	 */
	for (; read(fd, &c, 1) == 1; ) {
		debug((DEBUG_OBESE, "dgetln read %c", c));
		buffer[len++] = c;
		if (c == '\r') {
			got_cr = 1;
			continue;
		} else if (c == '\n') {
			/*
			 * HTTP/1.1 spec says to ignore CR and treat
			 * LF as the real line terminator.  even though
			 * the same spec defines CRLF as the line
			 * terminator, it is recommended in section 19.3
			 * to do the LF trick for tolerance.
			 */
			if (got_cr)
				len -= 2;
			else
				len -= 1;
			break;
		}

		if (len == buflen) {
			buflen *= 2;
			nbuffer = realloc(buffer, buflen);
			if (nbuffer == NULL) {
				free(buffer);
				buflen = 0;
				buffer = NULL;
				return NULL;
			}
			buffer = nbuffer;
		}
	}
	buffer[len] = '\0';
	debug((DEBUG_OBESE, "dgetln returns: ``%s'' with len %d", buffer, len));
	*lenp = len;
	return (buffer);
}
